package org.oasisopen.sca;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.lang.reflect.Field;
import java.net.URI;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Properties;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import com.helloservice.HelloService;
import com.vendor1.scaimpl.Vendor1SCAClientImpl;
import com.vendor2.scaimpl.Vendor2SCAClientImpl;
import com.vendor3.scaimpl.Vendor3SCAClientFactoryImpl;
import com.vendor3.scaimpl.Vendor3SCAClientImpl;
import com.vendor4.scaimpl.Vendor4SCAClientFactoryImpl;


public class TestSCAClientGetService {

    private static final String SCA_CLIENT_FACTORY_PROPERTY_NAME = SCAClientFactory.class.getName();
    
    private String serviceURI;
    private URI domainURI;
    private String originalSCAClientFactorySystemPropertyValue;

    @Before
    public void setup() throws Exception {
        serviceURI = "SomeHelloServiceURI";
        domainURI = new URI("SomeDomainURI");
        
        originalSCAClientFactorySystemPropertyValue = System.getProperty(SCA_CLIENT_FACTORY_PROPERTY_NAME);
        injectDefaultFactoryIntoSCAClientFactory(null);
    }
    
    @After
    public void destroy() throws Exception {
        if (originalSCAClientFactorySystemPropertyValue == null) {
            System.getProperties().remove(SCA_CLIENT_FACTORY_PROPERTY_NAME);
        } else {
            System.setProperty(SCA_CLIENT_FACTORY_PROPERTY_NAME, originalSCAClientFactorySystemPropertyValue);
        }
        injectDefaultFactoryIntoSCAClientFactory(null);
    }
    
    @Test
    public void testDefaultSCAClientFactoryImpl() throws NoSuchServiceException, NoSuchDomainException {
        final SCAClient scaClient = SCAClientFactory.newInstance();
        validateSCAClientInstance(scaClient, Vendor1SCAClientImpl.class);
    }
    
    @Test
    public void testSystemPropertyForVendor2SCAClientFactoryImpl() throws NoSuchServiceException, NoSuchDomainException {
        
        System.setProperty(SCA_CLIENT_FACTORY_PROPERTY_NAME, "com.vendor2.scaimpl.Vendor2SCAClientFactoryImpl");
        
        final SCAClient scaClient = SCAClientFactory.newInstance();
        validateSCAClientInstance(scaClient, Vendor2SCAClientImpl.class);
    }

    
    @Test
    public void testPropertyForVendor3SCAClientFactoryImpl() throws NoSuchServiceException, NoSuchDomainException {
        
        final Properties props = createPropertiesWithSCAFactoryImplName(Vendor3SCAClientFactoryImpl.class.getName());
        
        final SCAClient scaClient = SCAClientFactory.newInstance(props);

        validateSCAClientInstance(scaClient, Vendor3SCAClientImpl.class);
    }
    
    @Test
    public void testPropertyForNoSuchSCAClientFactoryImpl() throws NoSuchServiceException, NoSuchDomainException {
        final String missingClassName = "this.class.does.not.exist";
        final Properties props = createPropertiesWithSCAFactoryImplName(missingClassName);
        
        try {
            SCAClientFactory.newInstance(props);
            fail("Should have thrown an exception");
        } catch (SCARuntimeException ex) {
            assertTrue(ex.getMessage().indexOf(missingClassName) != -1);
        }
    }

    @Test
    public void testPropertyForNotSCAClientFactoryImpl() throws NoSuchServiceException, NoSuchDomainException {
        final String missingClassName = "java.lang.Object";
        final Properties props = createPropertiesWithSCAFactoryImplName(missingClassName);
        
        try {
            SCAClientFactory.newInstance(props);
            fail("Should have thrown an exception");
        } catch (SCARuntimeException ex) {
            assertTrue(ex.getMessage().indexOf(missingClassName) != -1);
        }
    }
    
    
    @Test
    public void testNoImplementationFound() {
        final ClassLoader emptyClassLoader = new URLClassLoader(new URL[0], null);
        
        try {
            SCAClientFactory.newInstance(emptyClassLoader);
            fail("Should have thrown an exception");
        } catch (SCARuntimeException ex) {
            assertEquals(ex.getMessage(), "Failed to find implementation for SCAClientFactory");
        }
    }
    
    @Test
    public void testUsingClassLoader() throws NoSuchServiceException, NoSuchDomainException {

        // TODO: This test probably needs enhancing
        final ClassLoader classLoader = Thread.currentThread().getContextClassLoader();
        final SCAClient scaClient = SCAClientFactory.newInstance(classLoader);
        validateSCAClientInstance(scaClient, Vendor1SCAClientImpl.class);
    }
    
    @Test
    public void testUsingInjectedVendor3Implementation() throws NoSuchServiceException, NoSuchDomainException, SecurityException, NoSuchFieldException, IllegalArgumentException, IllegalAccessException {

        final SCAClientFactory vendor3 = new Vendor3SCAClientFactoryImpl();
        injectDefaultFactoryIntoSCAClientFactory(vendor3);
        
        final SCAClient scaClient = SCAClientFactory.newInstance();
        validateSCAClientInstance(scaClient, Vendor3SCAClientImpl.class);
    }

    @Test
    public void testVendorSCAClientFactoryImplThrowsException() throws NoSuchServiceException, NoSuchDomainException {
        final String vendorImplClassName = Vendor4SCAClientFactoryImpl.class.getName();
        final Properties props = createPropertiesWithSCAFactoryImplName(vendorImplClassName);
        
        try {
            SCAClientFactory.newInstance(props);
            fail("Should have thrown an exception");
        } catch (SCARuntimeException ex) {
            assertTrue(ex.getMessage().indexOf(vendorImplClassName) != -1);
        }
    }

    private Properties createPropertiesWithSCAFactoryImplName(
            final String missingClassName)
    {
        final Properties props = new Properties();
        props.setProperty(SCA_CLIENT_FACTORY_PROPERTY_NAME, missingClassName);
        return props;
    }
    
    private void injectDefaultFactoryIntoSCAClientFactory(
            final SCAClientFactory vendor3) throws NoSuchFieldException,
            IllegalAccessException {

        final Field defaultFactoryField = SCAClientFactory.class.getDeclaredField("defaultFactory");
        defaultFactoryField.setAccessible(true);
        defaultFactoryField.set(null, vendor3);
        defaultFactoryField.setAccessible(false);
    }

    
    private void validateSCAClientInstance(final SCAClient scaClient, final Class<? extends SCAClient> expectedSCAClientImplClass)
        throws NoSuchServiceException, NoSuchDomainException {
        
        assertNotNull(scaClient);
        assertEquals(scaClient.getClass().getName(), expectedSCAClientImplClass.getName());
        
        final HelloService helloService = scaClient.getService(HelloService.class, serviceURI, domainURI);
        assertNotNull(helloService);
        
        final String reply = helloService.sayHello("Mark");
        assertEquals("Hello Mark", reply);
    }
    
}
